
/*
 * Fake implementations of some Linux kernel functions and structures.
 * Needed for testing in user space.
 */

#define spin_lock_init(unused) ;
#define spin_lock_bh(unused) ;
#define spin_unlock_bh(unused) ;
#define spinlock_t int

#define kmalloc(a,b) malloc(a)
#define kfree(a) free(a)
#define GFP_ATOMIC 0
#define GFP_KERNEL 0

#define ENOMEM 1

#define min(a,b) (((a) < (b)) ? (a) : (b))

struct list_head {
        struct list_head *next, *prev;
};

#define LIST_POISON1  ((void *) 0x00100100)
#define LIST_POISON2  ((void *) 0x00200200)

#define LIST_HEAD_INIT(name) { &(name), &(name) }

#define LIST_HEAD(name) \
        struct list_head name = LIST_HEAD_INIT(name)

#define INIT_LIST_HEAD(ptr) do { \
        (ptr)->next = (ptr); (ptr)->prev = (ptr); \
} while (0)

static inline void __list_add(struct list_head *new,
                              struct list_head *prev,
                              struct list_head *next)
{
        next->prev = new;
        new->next = next;
        new->prev = prev;
        prev->next = new;
}

static inline void list_add(struct list_head *new, struct list_head *head)
{
        __list_add(new, head, head->next);
}

static inline void __list_del(struct list_head * prev, struct list_head * next)
{
        next->prev = prev;
        prev->next = next;
}

static inline void list_del(struct list_head *entry)
{
        __list_del(entry->prev, entry->next);
        entry->next = LIST_POISON1;
        entry->prev = LIST_POISON2;
}

#define list_for_each(pos, head) \
        for (pos = (head)->next; pos->next, pos != (head); \
                pos = pos->next)


#define list_for_each_safe(pos, n, head) \
        for (pos = (head)->next, n = pos->next; pos != (head); \
                pos = n, n = pos->next)
