#define _GNU_SOURCE

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>

#include <linux/socket.h> 
#include <linux/types.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <linux/netlink.h>

#include <net/if.h>
#include <openssl/ripemd.h>

#include "../src/pla_netlink.h"
#include "../src/configs.h"

#include "ec.h"

#define MAX_PAYLOAD 1024  /* maximum payload size */

#define PLA_TTP_CONF_FILE "/etc/pla.conf"
#define PLA_CLIENT_CONF_FILE "/etc/pla_client.conf"
#define PLA_PRIVATE_KEY_FILE "/etc/pla_private.conf"

struct key_info
{
	elem_t rho;
	uint8_t b8;
	char id_hash[PLA_HASH_DIGEST_LENGTH];
	ecpoint_t e_y;
};

struct cert_info
{
	uint16_t deleg;                /* Delegation bits */
	uint16_t rights;               /* Rights bits */

	time_t not_before;             /* Validity time limits */
	time_t not_after;

	mpz_t ttp_sigma;               /* Used by the subject when operating as a TTP */
	mpz_t sigma;

	char r[PLA_HASH_DIGEST_LENGTH];
	mpz_t s;
};


static int netlink_fd;

FILE *output_file = NULL;
FILE *certificate_file = NULL;
FILE *key_file = NULL;

void key_info_init(struct key_info *ki)
{
	memset(ki, 0, sizeof(struct key_info));
//	mpz_init(ki->sigma);

	return;
}

void key_info_uninit(struct key_info *ki)
{
//	mpz_clear(ki->sigma);

	return;
}

void cert_info_init(struct cert_info *si)
{
	memset(si, 0, sizeof(struct cert_info));
	mpz_init(si->s);

	mpz_init(si->sigma);
	mpz_init(si->ttp_sigma);

	return;
}

void cert_info_uninit(struct cert_info *si)
{
	mpz_clear(si->ttp_sigma);
	mpz_clear(si->sigma);

	mpz_clear(si->s);

	return;
}


static char* string_hexread(char *source, unsigned char *target, unsigned int len)
{
	unsigned char *orig = target;
	unsigned remaining = len;

	while (remaining--) {
		sscanf(source, "%02x", target);
		target += 1;
		source += 2;
	}

	return source;
}

static void hexdump(FILE *file, unsigned char *buf, unsigned int len)
{
        while (len--)
                fprintf(file, "%02x", *buf++);
}

static int shexdump(char *target, unsigned char *buf, unsigned int len)
{
	char *orig = target;

        while (len--) {
                target += sprintf(target, "%02x", *buf++);
	}

	target += sprintf(target, "\n");

	return target - orig;
}

void nlh_fill(struct nlmsghdr* nlh, int type, int len, int flags)
{
	nlh->nlmsg_len = NLMSG_SPACE(len);
	nlh->nlmsg_pid = getpid();
	nlh->nlmsg_type = type;
	nlh->nlmsg_flags = flags;

	return;
}

int netlink_socket(void)
{
	int sock_fd;
	struct sockaddr_nl src_addr;
		
	sock_fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_PLA);
	if (sock_fd < 0) {
		perror("socket");
		exit(EXIT_FAILURE);
	}

	memset(&src_addr, 0, sizeof(struct sockaddr_nl));
	src_addr.nl_family = AF_NETLINK;      
	src_addr.nl_pid = getpid(); 
	src_addr.nl_groups = 0;

	if (bind(sock_fd, (struct sockaddr*) &src_addr, sizeof(src_addr)) < 0) {
		perror("bind");
		exit(EXIT_FAILURE);
	}
	
	return sock_fd;
}

int send_netlink(void *packet, int len, int type)
{
	struct sockaddr_nl dest_addr;

	struct nlmsghdr *nlh = NULL;
	struct iovec iov;
	struct msghdr msg;

	memset(&dest_addr, 0, sizeof(dest_addr));
	dest_addr.nl_family = AF_NETLINK;
	dest_addr.nl_pid = 0;  
	dest_addr.nl_groups = 0;

	nlh = malloc(NLMSG_SPACE(MAX_PAYLOAD));
	if (!nlh) {
		fprintf(stderr, "Out of memory in placonf.\n");
		exit(EXIT_FAILURE);
	}

	nlh_fill(nlh, type, MAX_PAYLOAD, 0);

	memcpy(NLMSG_DATA(nlh), packet, len);

	if (sendto(netlink_fd, nlh, nlh->nlmsg_len, 0, (struct sockaddr*) &dest_addr, sizeof(dest_addr)) < 0) {
		perror("sendto");
		exit(EXIT_FAILURE);
	}

	return 0;
}

int jlu_elem_to_hex(char *dst, elem_t src)
{
	int i;
	char *tmp_ptr = dst;

	for (i=0; i < NUMWORDS; i++)
		tmp_ptr += sprintf(tmp_ptr, "%08lx ",src[i]);

	tmp_ptr += sprintf(tmp_ptr, "\n");

	return tmp_ptr - dst;
}

int jlu_hex_to_elem(FILE *file, elem_t *dst)
{
	int i;
	uint32_t *ptr = (uint32_t*) &dst[0];

	for (i=0; i < NUMWORDS; i++) {
		if (fscanf(file, "%08lx", &ptr[i]) < 0)
			return -1;
	}

	return 0;
}

char* string_hex_to_elem(char *source, elem_t *dst)
{
	char *cur = source;
	int i;
	uint32_t *ptr = (uint32_t*) &dst[0];

	for (i=0; i < NUMWORDS; i++) {
		int add;

		add = sscanf(cur, "%08lx", &ptr[i]);
		if (add < 0) {
			fprintf(stderr, "Failure in hex_to_elem\n");
			exit(EXIT_FAILURE);
		}
		cur += 8+1;   /* Each field is 8+1 bytes long */
	}

	return cur;
}

void read_node_state_stream_s(FILE *file, struct key_info *issuer, struct key_info *subject, struct cert_info *ci)
{
	char tmp[2000];
	char *ptr = tmp;
	char *cur;

	memset(tmp, 0, sizeof(tmp));
	fread(tmp, sizeof(tmp), 1, file);

	strsep(&ptr, "\"");  /* Skip headers */

	cur = strsep(&ptr, "\"");
	cur = string_hex_to_elem(cur, &issuer->e_y.x) + 1;
	cur = string_hex_to_elem(cur, &issuer->e_y.y) + 1;
	cur = string_hex_to_elem(cur, &issuer->e_y.z) + 1;

	cur = string_hexread(cur, issuer->id_hash, PLA_HASH_DIGEST_LENGTH) + 1;
	cur = string_hex_to_elem(cur, &issuer->rho) + 1;

	int tmpb;
	sscanf(cur, "%d\n", &tmpb);
	issuer->b8 = tmpb;
	cur += 2;

	cur = strsep(&ptr, "\"");  /* Locator */
	cur = strsep(&ptr, "\"");  /* Locator value */

	/* Subject starts */
	cur = strsep(&ptr, "\"");  /* Pubkey */
	cur = strsep(&ptr, "\"");  /* Pubkey value */

	cur = string_hex_to_elem(cur, &subject->e_y.x) + 1;
	cur = string_hex_to_elem(cur, &subject->e_y.y) + 1;
	cur = string_hex_to_elem(cur, &subject->e_y.z) + 1;

	cur = string_hexread(cur, subject->id_hash, PLA_HASH_DIGEST_LENGTH) + 1;
	cur = string_hex_to_elem(cur, &subject->rho) + 1;

	sscanf(cur, "%d\n", &tmpb);
	subject->b8 = tmpb;
	cur += 2;

	/* Deleg, time, signature starts */
	cur = strsep(&ptr, "\"");  

	struct tm tm_nb;
	if (!strptime(ptr, "%Y-%m-%d %H:%M:%S", &tm_nb)) {
		fprintf(stderr, "Invalid time for not-before: %s\n", ptr);
		exit(EXIT_FAILURE);
	}
	ci->not_before = mktime(&tm_nb);

	cur = strsep(&ptr, "\"");  
	cur = strsep(&ptr, "\"");  

	struct tm tm_na;
	if (!strptime(ptr, "%Y-%m-%d %H:%M:%S", &tm_na)) {
		fprintf(stderr, "Invalid time for not-after: %s\n", ptr);
		exit(EXIT_FAILURE);
	}
	ci->not_after = mktime(&tm_na);

	cur = strsep(&ptr, "\"");  
	cur = strsep(&ptr, "\"");  /* rights */
	cur = strsep(&ptr, "\"");  /* rights value */

	{
		unsigned tmp_rights;
		sscanf(cur, "%x", &tmp_rights);
		ci->rights = tmp_rights;
	}


	cur = strsep(&ptr, "\"");  /* deleg */
	cur = strsep(&ptr, "\"");  /* deleg value */

	{
		unsigned tmp_deleg;
		sscanf(cur, "%x", &tmp_deleg);
		ci->deleg = tmp_deleg;
	}

	strsep(&ptr, "\"");
	cur = strsep(&ptr, "\"");
	mpz_set_str(ci->sigma, cur, 16);

	cur = strsep(&ptr, "\"");  /* ttp priv key */
	cur = strsep(&ptr, "\"");  /* ttp priv key value */
	mpz_set_str(ci->ttp_sigma, cur, 16);

	cur = strsep(&ptr, "\"");  /* signature value */
	cur = strsep(&ptr, "\"");  /* signature value */

	char *tmpcur = cur;
	tmpcur = string_hexread(tmpcur, ci->r, PLA_HASH_DIGEST_LENGTH);
	mpz_set_str(ci->s, tmpcur, 16);

	return;
}

int ecpoint_t_sprintf(char *cp, ecpoint_t *point)
{
	char *cp_orig = cp;

	cp += jlu_elem_to_hex(cp, point->x);

	cp += jlu_elem_to_hex(cp, point->y);

	cp += jlu_elem_to_hex(cp, point->z);

	return cp - cp_orig;
}

void id_create(char *id_hash)
{
	int i;

	/* jlu XXX: Yes, this does it in the dumbest possible way. */
	srandom(time(NULL));
	for (i=0; i<PLA_HASH_DIGEST_LENGTH; i++)
		id_hash[i] = random();

	return;
}

void fprint_public_key(FILE *file, struct key_info *node)
{
	char tmp[1000];
	char *fp = tmp;

	fp += ecpoint_t_sprintf(fp, &node->e_y);
	fp += shexdump(fp, node->id_hash, PLA_HASH_DIGEST_LENGTH);
	fp += jlu_elem_to_hex(fp, node->rho);
	fp += sprintf(fp, "%d", node->b8);

	fprintf(file, "\"%s\"\n", tmp);

	return;
}

void fprint_private_key(FILE *file, struct cert_info *node)
{
	char *str = mpz_get_str(NULL, 16, node->sigma);
	if(!str) {
		fprintf(stderr, "Out of memory on fprint_private_key\n");
		exit(EXIT_FAILURE);
	}
	fprintf(file, "%s", str);
	free(str);

	return;
}

void write_node_state_s(FILE *file, struct key_info *issuer, struct key_info *subject, struct cert_info *ci)
{
	fprintf(file, "(cert \n");

	fprintf(file, " (issuer\n");
	fprintf(file, "  (pub-key\n");
	fprint_public_key(file, issuer);
	fprintf(file, "   locator ");
	fprintf(file, "\"NULL\"");
	fprintf(file, "\n");
	fprintf(file, "  )\n");
	fprintf(file, " )\n");

	fprintf(file, " (subject\n");
	fprintf(file, "  (pub-key\n");
	fprint_public_key(file, subject);
	fprintf(file, "  )\n");
	fprintf(file, " )\n");

	char nb[50];
	char na[50];
	struct tm tm_nb;
	struct tm tm_na;

	localtime_r(&ci->not_before, &tm_nb);
	localtime_r(&ci->not_after, &tm_na);

	strftime(nb, sizeof(nb), "%Y-%m-%d %H:%M:%S", &tm_nb);
	strftime(na, sizeof(na), "%Y-%m-%d %H:%M:%S", &tm_na);

	fprintf(file, " (valid not-before \"%s\" not-after \"%s\")\n", nb, na);
	fprintf(file, " (rights \"%x\")\n", ci->rights);
	fprintf(file, " (deleg \"%x\")\n", ci->deleg);

	fprintf(file, " (private-key \"");
	fprint_private_key(file, ci);
	fprintf(file, "\")\n");

	fprintf(file, " (ttp-private-key \"");
	{
		char *str = mpz_get_str(NULL, 16, ci->ttp_sigma);
		if(!str) {
			fprintf(stderr, "Out of memory on ttp_sigma print\n");
			exit(EXIT_FAILURE);
		}
		fprintf(file, "%s", str);
		free(str);
	}
	fprintf(file, "\")\n");

	fprintf(file, " (signature\n\"");
	hexdump(file, ci->r, PLA_HASH_DIGEST_LENGTH);	
	fprintf(file, "\n");
	mpz_out_str(file, 16, ci->s);
	fprintf(file, "\"\n )\n");

	fprintf(file, ")\n");

	return;
}

void create_key(elem_t *rho, mpz_t *sigma, uint8_t *b8, char *client_id_hash, mpz_t *ttp_private)
{
	/* Create new key for the client */
	mpz_t k;
	mpz_t rhoi;
	ecpoint_t kgd;
	
	mpz_init(k);
	ec_keygen(&kgd,k);
	
	int b;
	
	ec_compress_point(&b,&kgd);
	*b8 = b;
	elem_t hid;
	elem_import_bytes(hid, client_id_hash); 
	elem_chomp(hid); /* hash of the ID */
	gnb_add(*rho,hid,kgd.x);
	
	mpz_init(rhoi);
	mpz_init(*sigma);
	elem_to_int(rhoi,*rho);
	mpz_fdiv_r(rhoi,rhoi,e.r);
	mpz_mul(*sigma,rhoi,*ttp_private);
	mpz_add(*sigma,k,*sigma);
	mpz_fdiv_r(*sigma,*sigma,e.r);
	
	return;
}

void set_userspace_target(char *target)
{
	struct pla_netlink_target userspace_target;

	memset(&userspace_target, 0, sizeof(userspace_target));

	if (inet_pton(AF_INET, target, &userspace_target.addr) <= 0) {
		fprintf(stderr, "Error in address parameter\n");
		exit(EXIT_FAILURE);
	}

	send_netlink(&userspace_target, sizeof(userspace_target), NLMSG_PLA_ADD );

	return;
}

static int cert_to_bin(char *target, int target_len, struct key_info *issuer, struct key_info *subject, struct cert_info *ci)
{
	char *cur = target;
	uint16_t *tmp16;
	uint32_t *tmp32;

	memset(target, 0, target_len);
	cur += key_info_to_bin(cur, issuer);
	cur += key_info_to_bin(cur, subject);


	/* Convert delegation bits */
	tmp16 = ((uint16_t *) cur);
	*tmp16 = htons(ci->deleg);
	tmp16++;
	cur = ((char *) tmp16);

	/* Convert not-before */
	tmp32 = ((uint32_t *) cur);
	*tmp32 = htonl(ci->not_before);
	tmp32++;
	cur = ((char *) tmp32);

	/* Not-after */
	tmp32 = ((uint32_t *) cur);
	*tmp32 = htonl(ci->not_after);
	tmp32++;
	cur = ((char *) tmp32);

	return cur - target;
}

void cert_to_digest(char *digest, struct key_info *issuer, struct key_info *subject, struct cert_info *ci)
{
	char bin_cert[1000];
	int len = cert_to_bin(bin_cert, sizeof(bin_cert), issuer, subject, ci);

	RIPEMD160(bin_cert, len, digest);

	return;
}

int set_interfaces(char **argv, int argc, int optind)
{
	int if_in, if_out;
	char* policystr;
	int policynum;

	if_in = if_nametoindex(argv[optind]);
	if (if_in == 0) {
		fprintf(stderr, "No such interface: %s\n", argv[optind]);
		exit(EXIT_FAILURE);
	}
	optind++;

	if_out = if_nametoindex(argv[optind]);
	if (if_out == 0) {
		fprintf(stderr, "No such interface: %s\n", argv[optind]);
		exit(EXIT_FAILURE);
	}
	optind++;

	policystr = argv[optind];
	if (strcmp(policystr, "drop") == 0)
		policynum = PLA_FORWARD_POLICY_DROP;
	else if (strcmp(policystr, "add") == 0)
		policynum = PLA_FORWARD_POLICY_SIGN;
	else if (strcmp(policystr, "verify") == 0)
		policynum = PLA_FORWARD_POLICY_VERIFY;
	else if (strcmp(policystr, "strip") == 0)
		policynum = PLA_FORWARD_POLICY_STRIP;
	else {
		fprintf(stderr, "Invalid policy string: %s\n",  policystr);
		exit(EXIT_FAILURE);
	}

	if (if_out == 0 && strcmp(argv[optind], "local") != 0) {
		fprintf(stderr, "No such interface: %s\n", argv[optind]);
		exit(EXIT_FAILURE);
	}
	optind++;

	fprintf(stderr, "Setting policy: %d\n", policynum);

	struct pla_netlink_set_interface if_set = {
		.if_in  = if_in,
		.if_out = if_out,
		.policy = policynum,
		};

	
	fprintf(stderr, "Indexes: %d %d\n", if_set.if_in, if_set.if_out);

	send_netlink(&if_set, sizeof(if_set), NLMSG_PLA_SET_INTERFACES );

	return 0;
}

void send_cert_netlink(ecpoint_t *e_y)
{
        
	/* Send acceptable public key to kernel over netlink */
	struct pla_netlink_add_src *src;
	int src_len = sizeof(struct pla_netlink_add_src) + sizeof(ecpoint_t);
	
	src = malloc(src_len);
	if (!src) {
		fprintf(stderr, "Out of memory in send_cert_netlink\n");
		exit(EXIT_FAILURE);
	}
	
	src->publen = sizeof(ecpoint_t);
	char *tmp = &src->data[0];
	
	/* TTP public key */
	memcpy(tmp, e_y, sizeof(ecpoint_t));   
	tmp += src->publen;
	
	send_netlink(src, src_len, NLMSG_PLA_ADD_CERT);
	
	free(src);

	return;
}

int add_cert(char **argv, int argc, int optind)
{
	struct key_info issuer, subject;
	struct cert_info si;
	
	key_info_init(&issuer);
	key_info_init(&subject);
	cert_info_init(&si);
	
	if (!certificate_file) {
		fprintf(stderr, "No certificate file set\n");
		exit(EXIT_FAILURE);
	}
	
	read_node_state_stream_s(certificate_file, &issuer, &subject, &si);
	
	send_cert_netlink(&subject.e_y);
	
	cert_info_uninit(&si);
	key_info_uninit(&subject);
	key_info_uninit(&issuer);

	return 0;
}

void set_time_in_ci(struct cert_info *ci, char *not_before, char *not_after)
{
	struct tm tm_na;  /* Not after */
	struct tm tm_nb;  /* Not before */

	if (!strptime(not_before, "%Y-%m-%d %H:%M:%S", &tm_nb)) {
		fprintf(stderr, "Invalid time for not-before: %s\n", not_before);
		exit(EXIT_FAILURE);
	}

	if (!strptime(not_after, "%Y-%m-%d %H:%M:%S", &tm_na)) {
		fprintf(stderr, "Invalid time for not-before: %s\n", not_after);
		exit(EXIT_FAILURE);
	}

	ci->not_before = mktime(&tm_nb);
	ci->not_after = mktime(&tm_na);

	return;
}


int ttp_create(char **args, int argc, int optind)
{
	struct key_info issuer;
	struct key_info subject;
	struct cert_info ci;

	key_info_init(&issuer);
	key_info_init(&subject);
	cert_info_init(&ci);

	char digest[RIPEMD160_DIGEST_LENGTH + 1]; // jlu XXX: Workaround for EC bug.
	digest[RIPEMD160_DIGEST_LENGTH] = 0;      // jlu XXX: Workaround for EC bug.

	/* Create a new TTP */
	mpz_t sig_sigma;
	mpz_init(sig_sigma);

	ec_keygen(&(e.y), ci.sigma); /* e.y is domain public key of TTP */
	/* e_y, sigma */
	
	id_create(subject.id_hash);
	create_key(&subject.rho, &sig_sigma, &subject.b8, subject.id_hash, &ci.sigma);
	/* e_y, sigma, id_hash -> rho, sig_sigma, b8 */

	memcpy(&subject.e_y, &e.y, sizeof(subject.e_y));
	memcpy(&issuer.e_y, &e.y, sizeof(issuer.e_y));

#if 0
	{
		char new_id[20];
		elem_t new_rho;
		uint8_t new_b8;
		mpz_t new_sigma;
		ecselfsig_t new_sig;
		char new_hash[20+1];

		memset(new_hash, 0, sizeof(new_hash)); new_hash[2]=0x23; new_hash[10]=0xAA;
		memset(&new_sig, 0, sizeof(new_sig));
		memset(new_rho, 0, sizeof(new_rho));
		mpz_init(new_sigma);
		new_b8=0;

		// Create a new id
		id_create(new_id);

		create_key(&new_rho, &new_sigma, &new_b8, new_id, &sig_sigma);

		// sign something
		new_sig.b = new_b8;
		memcpy(&new_sig.rho, &new_rho, sizeof(new_rho));
		ec_sign_self(&new_sig, new_hash, new_sigma);

		// verify the signature
		fprintf(stderr, "Verify: %d\n", ec_verify_self(&new_sig, new_hash, new_id));
		
	}
#endif

	/* Sign it. */
	ecselfsig_t sig;
	
	sig.b = subject.b8;
	memcpy(&sig.rho, &subject.rho, sizeof(subject.rho));
	
	ci.deleg = 0xFFFF;
	ci.rights = 0xFFFF;

	set_time_in_ci(&ci, args[optind], args[optind+1]);
	optind += 2;

#if 1
	// jlu XXX: this should not be given to the node. The node can use it to forge
	// certificates. Fix this!!!!
	mpz_set(ci.ttp_sigma, sig_sigma);
#endif


	cert_to_digest(digest, &issuer, &subject, &ci);
	ec_sign_self(&sig, digest, sig_sigma);
	
	mpz_set(ci.s, sig.s); 
	memcpy(ci.r, sig.r, PLA_HASH_DIGEST_LENGTH);
	
	/* Subject and subject are the same. No need to make a copy */
	write_node_state_s(output_file, &issuer, &subject, &ci);

	mpz_clear(sig_sigma);
	
	key_info_uninit(&subject);
	key_info_uninit(&issuer);
	cert_info_uninit(&ci);

	return 0;
}


int cert_verify(void)
{
	int ret = -1;

	struct stat sb;

	struct key_info issuer;
	struct key_info subject;
	struct cert_info si;

	key_info_init(&issuer);
	key_info_init(&subject);
	cert_info_init(&si);

	char digest[RIPEMD160_DIGEST_LENGTH + 1]; // jlu XXX: Workaround for EC bug.
	digest[RIPEMD160_DIGEST_LENGTH] = 0;      // jlu XXX: Workaround for EC bug.

	if (!certificate_file) {
		fprintf(stderr, "No certificate file name given.\n");
		exit(EXIT_FAILURE);
	}
	read_node_state_stream_s(certificate_file, &issuer, &subject, &si);
	memcpy(&e.y, &issuer.e_y, sizeof(e.y));   

	/* Check not_before and not_after */
	if (si.not_before > time(NULL) || si.not_after < time(NULL)) {
		fprintf(stderr, "Failed verify: Time.\n");
		exit(2);
	}
	
	/* Verify it. */
	ecselfsig_t sig;
	
	sig.b = subject.b8;
	memcpy(&sig.rho, &subject.rho, sizeof(subject.rho));
	
	mpz_init(sig.s);
	mpz_set(sig.s, si.s); 
	memcpy(sig.r, si.r, PLA_HASH_DIGEST_LENGTH);
	
	cert_to_digest(digest, &issuer, &subject, &si);

	char subject_id_hash[PLA_HASH_DIGEST_LENGTH+1];
	subject_id_hash[PLA_HASH_DIGEST_LENGTH] = 0;
	memcpy(&subject_id_hash, &subject.id_hash, PLA_HASH_DIGEST_LENGTH);

#if 0
	fprintf(stderr, "\n");
	fprintf(stderr, "e.y.x: "); elem_print_hex(e.y.x);
	fprintf(stderr, "e.y.y: "); elem_print_hex(e.y.y);
	fprintf(stderr, "e.y.z: "); elem_print_hex(e.y.z);
	fprintf(stderr, "\n");
	hexdump(stderr, digest, 20);fprintf(stderr, "\n");
	hexdump(stderr, subject_id_hash, 20); fprintf(stderr, "\n");
	hexdump(stderr, (char*) &sig.rho, sizeof(sig.rho)); fprintf(stderr, " b: %d\n", sig.b);

	hexdump(stderr, sig.r, PLA_HASH_DIGEST_LENGTH); fprintf(stderr, "\n");
	mpz_out_str(stderr, 16, sig.s);fprintf(stderr, "\n");

#endif

	if(ec_verify_self(&sig, digest, subject_id_hash) == 1) {
		ret = 1;
		fprintf(stderr, "Successful verify\n");
	} else {
		ret = 2;
		fprintf(stderr, "Failed verify: Signature.\n");
	}

	key_info_uninit(&subject);
	key_info_uninit(&issuer);
	cert_info_uninit(&si);

	return ret;
}


int id_sign(char **args, int argc, int optind)
{
	char digest[RIPEMD160_DIGEST_LENGTH];

	struct cert_info signature_null;
	struct cert_info input_cert;
	struct cert_info ci;
	
	struct key_info tmp_null;
	struct key_info subject;
	struct key_info issuer;

	key_info_init(&tmp_null);
	key_info_init(&subject);
	key_info_init(&issuer);
	cert_info_init(&signature_null);
	cert_info_init(&input_cert);
	cert_info_init(&ci);

	mpz_t sig_sigma;
	mpz_init(sig_sigma);
	
	/* Read in the id hash */
	if (!key_file) {
		fprintf(stderr, "No key to be signed.\n");
		exit(EXIT_FAILURE);
	}
	read_node_state_stream_s(key_file, &tmp_null, &subject, &signature_null);

	if (!certificate_file) {
		fprintf(stderr, "No certificate file given. Trying " PLA_TTP_CONF_FILE "\n");
		certificate_file = fopen(PLA_TTP_CONF_FILE, "r");
		if (!certificate_file) {
			fprintf(stderr, "Error opening file\n");
			exit(EXIT_FAILURE);
		}
	}
	read_node_state_stream_s(certificate_file, &tmp_null, &issuer, &input_cert);
	memcpy(&e.y, &issuer.e_y, sizeof(e.y));

	mpz_t issuer_sigma;
	mpz_init(issuer_sigma);
	mpz_set(issuer_sigma, input_cert.sigma);
	/* Zero out issuer private key. It's not needed anymore */
	mpz_set_ui(input_cert.sigma, 0);

	/* Create a new key for the node */
	create_key(&subject.rho, &sig_sigma, &subject.b8, subject.id_hash, &issuer_sigma);
	ec_keygen(&subject.e_y, ci.sigma);

	mpz_clear(issuer_sigma);

	memset(issuer.id_hash, 0, sizeof(issuer.id_hash));
	memset(issuer.rho, 0, sizeof(issuer.rho));
	issuer.b8 = 0;

	/* Rights */
	if (optind >= argc) {
		fprintf(stderr, "Rights parameter missing\n");
		exit(EXIT_FAILURE);
	}
	uint16_t tmp_rights = strtol(args[optind], NULL, 16);
	if (errno == ERANGE) {
		fprintf(stderr, "Invalid rights string\n");
		exit(EXIT_FAILURE);
	}
	ci.rights = tmp_rights;
	optind++;

	/* Delegation */
	if (optind >= argc) {
		fprintf(stderr, "Delegation parameter missing\n");
		exit(EXIT_FAILURE);
	}
	ci.deleg = strtol(args[optind], NULL, 16);
	if (errno == ERANGE) {
		fprintf(stderr, "Invalid delegation string\n");
		exit(EXIT_FAILURE);
	}
	optind++;

	/* Not before, not after */
	if (optind + 1 >= argc) {
		fprintf(stderr, "Date parameters missing.\n");
		exit(EXIT_FAILURE);
	}
	set_time_in_ci(&ci, args[optind], args[optind+1]);
	optind += 2;


	ecselfsig_t sig;

#if 1
	// jlu XXX: this should not be given to the node. The node can use it to forge
	// certificates. Fix this!!!!
	mpz_set(ci.ttp_sigma, sig_sigma);
#endif

	cert_to_digest(digest, &issuer, &subject, &ci);


	sig.b = subject.b8;
	memcpy(&sig.rho, &subject.rho, sizeof(subject.rho));
	ec_sign_self(&sig, digest, sig_sigma);
	

	mpz_set(ci.s, sig.s); 
	memcpy(ci.r, sig.r, PLA_HASH_DIGEST_LENGTH);



#if 0
	{
		/* Test the values: */
		  static byte_t id_hash[100];

		  fprintf(stderr, "\n");
		  fprintf(stderr, "e.y.x: "); elem_print_hex(e.y.x);
		  fprintf(stderr, "e.y.y: "); elem_print_hex(e.y.y);
		  fprintf(stderr, "e.y.z: "); elem_print_hex(e.y.z);
		  fprintf(stderr, "\n");
		  hexdump(stderr, digest, 20); fprintf(stderr, "\n");
		  hexdump(stderr, subject.id_hash, 20);fprintf(stderr, "\n");
		  hexdump(stderr, (char*) &sig.rho, sizeof(sig.rho)); fprintf(stderr, " b: %d\n", sig.b);

		  hexdump(stderr, sig.r, PLA_HASH_DIGEST_LENGTH); fprintf(stderr, "\n");
		  mpz_out_str(stderr, 16, sig.s);fprintf(stderr, "\n");

		  memcpy(id_hash, subject.id_hash, sizeof(subject.id_hash));

		  int ret = ec_verify_self(&sig,digest,id_hash);
		  fprintf(stderr, "Verify returned: %d\n", ret);
	}
#endif

	write_node_state_s(output_file, &issuer, &subject, &ci);

	mpz_clear(sig_sigma);

	cert_info_uninit(&ci);
	cert_info_uninit(&input_cert);
	cert_info_uninit(&signature_null);
	key_info_uninit(&issuer);
	key_info_uninit(&subject);		
	key_info_uninit(&tmp_null);
	
	return 0;
}

void send_id_netlink(ecpoint_t *e_y, mpz_t *sigma, elem_t *rho, int8_t *b, char *id_hash)
{
        
	/* Send certificate and private key to kernel over netlink */
	struct pla_netlink_add_src *src;
	int src_len = sizeof(struct pla_netlink_add_src) + sizeof(ecpoint_t) + PLA_PRIVKEY_LEN + PLA_CERT_LEN + PLA_HASH_DIGEST_LENGTH;
	
	src = malloc(src_len);
	if (!src) {
		fprintf(stderr, "Out of memory in send_id_netlink\n");
		exit(EXIT_FAILURE);
	}
	
	src->publen = sizeof(ecpoint_t);
	src->privlen = PLA_PRIVKEY_LEN;
	src->certlen = PLA_CERT_LEN;
	src->id_hashlen = PLA_HASH_DIGEST_LENGTH;
	
	char *tmp = &src->data[0];
	
	/* TTP public key */
	memcpy(tmp, e_y, sizeof(ecpoint_t));   
	tmp += src->publen;
	
	memset(tmp, 0xAB, src->privlen);  // jlu XXX: deadbeef
	ec_privkey_to_bin(tmp, sigma);
	tmp += PLA_PRIVKEY_LEN;
	
	memset(tmp, 0xCC, src->certlen);  // jlu XXX: deadbeef
	int clen = ec_cert_to_bin(tmp, *rho, *b);
	tmp += PLA_CERT_LEN;
	
	memcpy(tmp, id_hash, src->id_hashlen);
	tmp += src->id_hashlen;
	
	send_netlink(src, src_len, NLMSG_PLA_SET_SRC);
	
	free(src);

	return;
}

int key_info_to_bin(char *target, struct key_info *info)
{
	char *cur = target;

	// e_y
	memcpy(cur, &info->e_y, sizeof(info->e_y));
	cur += sizeof(info->e_y);

	// id_hash
	memcpy(cur, &info->id_hash, sizeof(info->id_hash));
	cur += sizeof(info->id_hash);

	// rho
	// b
	cur += ec_cert_to_bin(cur, info->rho, info->b8);

	return cur - target;
}

void kernelconfig(char **args, int argc, int optind)
{
	if (optind >= argc) {
		fprintf(stderr, "Insufficient parameters for command: kernelconfig.\n");
		exit(EXIT_FAILURE);
	}

	if (strcmp(args[optind], "help") == 0) {
		optind++;
		goto out_err;
	}

	if (strcmp(args[optind], "cryptotarget") == 0) {
		optind++;

		set_userspace_target(args[optind]);

		return;
	}

	if (strcmp(args[optind], "setinterface") == 0) {
		optind++;

		set_interfaces(args, argc, optind);
		return;
	}

	if (strcmp(args[optind], "addcert") == 0) {
		optind++;

		add_cert(args, argc, optind);
		return;
	}

	if (strcmp(args[optind], "configid") == 0) {
		optind++;

		struct key_info issuer, subject;
		struct cert_info ci;

		key_info_init(&issuer);
		key_info_init(&subject);
		cert_info_init(&ci);

		if (!certificate_file) {
			fprintf(stderr, "No certificate file set\n");
			exit(EXIT_FAILURE);
		}

		read_node_state_stream_s(certificate_file, &issuer, &subject, &ci);
		
		if (mpz_cmp_ui(ci.sigma, 0) == 0) {
			fprintf(stderr, "Error: Trying to configure kernel id without a private key.\n");
			exit(EXIT_FAILURE);
		}
		
#if 1
		fprintf(stderr, "Configuring rho: ");
		hexdump(stderr, (char*) &subject.rho, sizeof(subject.rho));
		fprintf(stderr, "\n");
#endif

		send_id_netlink(&issuer.e_y, &ci.ttp_sigma, &subject.rho, &subject.b8, subject.id_hash);
		
		cert_info_uninit(&ci);
		key_info_uninit(&subject);
		key_info_uninit(&issuer);

		return;
	}

	fprintf(stderr, "Unknown command after kernelconfig\n");
	goto out_err;

	return;

out_err:
	fprintf(stdout, "\nUsage: placonf [OPTIONS] kernelconfig {cryptotarget|setinterface|configid|help}\n"); 
	exit(EXIT_FAILURE);

	return;
}

int id(char **args, int argc, int optind)
{
	if (optind >= argc) {
		fprintf(stderr, "Insufficient parameters for command: id.\n");
		goto out_err;
	}

	if (strcmp(args[optind], "sign") == 0) {
		optind++;
		return id_sign(args, argc, optind);
	}

	if (strcmp(args[optind], "create") == 0) {
		optind++;

		struct key_info ki;
		struct cert_info ci;

		key_info_init(&ki);
		cert_info_init(&ci);

		id_create(ki.id_hash);

		write_node_state_s(output_file, &ki, &ki, &ci);

		cert_info_uninit(&ci);
		key_info_uninit(&ki);
		return 0;
	}

	fprintf(stderr, "Unknown command after id\n");

out_err:
	fprintf(stdout, "\nUsage: placonf [OPTIONS] id {create|sign}\n"); 

	return -1;
}

void chain(char **args, int argc, int optind)
{

//	*****


	return;
}


int cert(char **args, int argc, int optind)
{
	char id_hash[PLA_HASH_DIGEST_LENGTH];

	if (optind >= argc) {
		fprintf(stderr, "Insufficient parameters for command: cert.\n");
		goto out_err;
	}

	if (strcmp(args[optind], "verify") == 0) {
		optind++;
		return cert_verify();
	}

#if 0
	if (strcmp(args[optind], "chain") == 0) {
		optind++;
		chain(args, argc, optind);
		return;
	}
#endif

	fprintf(stderr, "Unknown command after cert.\n");
	goto out_err;

	return;

out_err:
	fprintf(stdout, "\nUsage: placonf [OPTIONS] cert verify\n"); 

	return -1;
}

int ttp(char **args, int argc, int optind)
{
	char id_hash[PLA_HASH_DIGEST_LENGTH];

	if (optind >= argc) {
		fprintf(stderr, "Insufficient parameters for command: ttp.\n");
		goto out_err;
	}

	if (strcmp(args[optind], "create") == 0) {
		optind++;
		return ttp_create(args, argc, optind);
	}

	fprintf(stderr, "Unknown command after ttp\n");
	goto out_err;

out_err:
	fprintf(stdout, "\nUsage: placonf [OPTIONS] ttp create\n"); 

	return -1;
}

int main(int argc, char **argv) 
{
	int ret = -1;

	int c;
	int use_hw = 0;

	output_file = stdout;

	netlink_fd = netlink_socket();
	if (netlink_fd < 0) {
		perror("placonf");
		exit(EXIT_FAILURE);
	}

	ec_init();

	while (1) {
		c = getopt_long (argc, argv, "o:C:K:H", NULL, NULL);
		if (c == -1)
			break;
		
		switch (c) {
		case 'H':
			use_hw = 1;
			break;

		case 'o':
			output_file = fopen(optarg, "w");
			if (!output_file) {
				perror("fopen");
				exit(EXIT_FAILURE);
			}
			break;

		case 'K':
			key_file = fopen(optarg, "r");
			if (!key_file) {
				perror("fopen");
				exit(EXIT_FAILURE);
			}
			break;

		case 'C':
			certificate_file = fopen(optarg, "r");
			if (!certificate_file) {
				perror("fopen");
				exit(EXIT_FAILURE);
			}
			break;

		case '?':
			exit(EXIT_FAILURE);
			
		default:
			fprintf (stderr, "?? getopt returned character code 0%o ??\n", c);
			exit(EXIT_FAILURE);

		}
	}
	
	if (use_hw) 
		ec_open();

	if (optind < argc) {

		if (strcmp(argv[optind], "help") == 0) {
			fprintf(stdout, "Usage: placonf [OPTIONS] COMMAND\n");
			fprintf(stdout, "Options:\n");
			fprintf(stdout, "         -o output_file\n");
			fprintf(stdout, "         -C certificate_file\n");
			fprintf(stdout, "         -K key_file\n");
			fprintf(stdout, "         -H    (Use crypto accelerator)\n");
			fprintf(stdout, "Commands:\n");
			fprintf(stdout, "         help | cert | ttp | id | kernelconfig\n");
			goto out;
		}

		if (strcmp(argv[optind], "cert") == 0) {
			++optind;
			cert(argv, argc, optind);
			goto out;
		}

		if (strcmp(argv[optind], "ttp") == 0) {
			++optind;
			ret = ttp(argv, argc, optind);
			goto out;
		}

		if (strcmp(argv[optind], "id") == 0) {
			++optind;
			ret = id(argv, argc, optind);
			goto out;
		}

		if (strcmp(argv[optind], "kernelconfig") == 0) {
			++optind;
			kernelconfig(argv, argc, optind);
			ret = 0;
			goto out;
		}

		fprintf(stderr, "Unknown command: %s\n", argv[optind]);
		exit(EXIT_FAILURE);
	}

out:
	close(netlink_fd);

	fclose(output_file);  /* This always exists */

	if (key_file)
		fclose(key_file);

	if (certificate_file)
		fclose(certificate_file);

	ec_close();

	return ret;
}    
